(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lib_print_xml.ml,v 1.13 2014/01/31 09:51:52 fclement Exp $ *)

(* Runtime library for pretty-print_xmlers generated with OCamlGen. *)

open Format;;

(* Printing tokens that must be escaped.
   These escapes are not in the sense of String.escaped or
   Char.escaped, since accented chars should not be replaced by the
   decimal encoding of their ascii code. *)

(* Print_xmling strings with quotes. *)
let escape_string s =
  let more = ref 0 in
  for i = 0 to String.length s - 1 do
   match s.[i] with
   | '\\' | '\"' | '\'' -> incr more
   |  _ -> ()
  done;
  if !more = 0 then s else
  let res = String.create (String.length s + !more) in
  let j = ref 0 in
  for i = 0 to String.length s - 1 do
   let c = s.[i] in
   match c with
   | '\\' | '\"' | '\'' ->
     res.[!j] <- '\\'; incr j;
     res.[!j] <- c; incr j
   | _ ->
     res.[!j] <- c; incr j
  done;
  res
;;

let escape_char c =
  if c = '\\' then "\\\\" else
  if c = '\'' then "\\\'" else
  if c = '\"' then "\\\"" else
  String.make 1 c
;;

let print_xml_quoted_string ppf s =
  fprintf ppf "\"%s\"" (escape_string s)
;;
let print_xml_quoted_char ppf c =
  fprintf ppf "'%s'" (escape_char c)
;;
let print_xml_quoted_int ppf i =
  if i < 0 then fprintf ppf "(%i)" i else fprintf ppf "%i" i
;;
let print_xml_quoted_nativeint ppf i =
  if i < 0n then fprintf ppf "(%ni)" i else fprintf ppf "%ni" i
;;
let print_xml_quoted_int32 ppf i =
  if i < 0l then fprintf ppf "(%li)" i else fprintf ppf "%li" i
;;
let print_xml_quoted_int64 ppf i =
  if i < 0L then fprintf ppf "(%Li)" i else fprintf ppf "%Li" i
;;
let print_xml_quoted_float ppf f =
  if f <= 0.0 then fprintf ppf "(%f)" f else fprintf ppf "%f" f
;;

(* Iterators *)
let print_xml_Couple pp_a pp_b ppf (a, b) =
  Format.fprintf ppf "@[<1>(%a,@ %a)@]" pp_a a pp_b b
;;

let print_xml_Triple pp_a pp_b pp_c ppf (a, b, c) =
  Format.fprintf ppf "@[<1>(%a,@ %a,@ %a)@]" pp_a a pp_b b pp_c c
;;

let print_xml_List f ppf l =
  let pl ppf l =
    List.iter (fun x -> fprintf ppf "%a;@ " f x) l in
  fprintf ppf "@[<2>[@ %a@,]@]" pl l
;;

let print_xml_Array f ppf v =
  let pl ppf v =
    let l = Array.length v in
    for i = 0 to l - 1 do
      fprintf ppf "%a;@ " f v.(i)
    done in
  fprintf ppf "@[<3>[|@ %a@,|]@]" pl v
;;

let print_xml_Option f ppf = function
  | None -> fprintf ppf "%s" "None"
  | Some x -> fprintf ppf "@[<2>Some@ %a@]" f x
;;

let print_xml_Ref f ppf x = fprintf ppf "@[<2>ref@ %a@]" f !x
;;

let print_xml_couple fmt pp_a pp_b ppf (a, b) =
  Format.fprintf ppf "%a" pp_a a;
  Format.fprintf ppf fmt;
  Format.fprintf ppf "%a" pp_b b
;;

let print_xml_triple fmt pp_a pp_b pp_c ppf (a, b, c) =
  Format.fprintf ppf "%a" pp_a a;
  Format.fprintf ppf fmt;
  Format.fprintf ppf "%a" pp_b b;
  Format.fprintf ppf fmt;
  Format.fprintf ppf "%a" pp_c c
;;

let print_xml_list fmt f =
  let rec loop ppf = function
    | [] -> ()
    | x :: l ->
      f ppf x;
      if l <> [] then
      begin Format.fprintf ppf fmt; loop ppf l end in
  loop
;;

let print_xml_array fmt f ppf v =
  let rec loop ppf i j =
    match compare j 1 with
    | 0 -> f ppf v.(i)
    | 1 -> f ppf v.(i); Format.fprintf ppf fmt; loop ppf (i + 1) (j - 1)
    | _ -> () in
  loop ppf 0 (Array.length v)
;;

let print_xml_option fmt f ppf = function
  | None -> ()
  | Some x ->
    Format.fprintf ppf fmt;
    f ppf x
;;

let print_xml_ref fmt f ppf x =
  Format.fprintf ppf fmt;
  f ppf !x
;;

let print_xml_unit ppf () = fprintf ppf "%s" "()"
;;

let print_xml_bool ppf = function
  | true -> fprintf ppf "%s" "true"
  | _ -> fprintf ppf "%s" "false"
;;

let print_xml_poly ppf _ =
  fprintf ppf "%s" "<poly>"
;;

let print_xml_abstract ppf _ = fprintf ppf "%s" "<abstract>"
;;

let print_xml_external = print_xml_abstract;;

let print_xml_fun _p_a _p_b ppf _f =
  fprintf ppf "%s" "<fun>"
;;
